"""LAN Switch Interconnect product block that has all parameters of a subscription throughout its lifecycle."""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle

from gso.products.product_blocks.router import RouterBlock, RouterBlockInactive, RouterBlockProvisioning
from gso.products.product_blocks.switch import SwitchBlock, SwitchBlockInactive, SwitchBlockProvisioning
from gso.utils.types.interfaces import LAGMemberList
from gso.utils.types.ip_address import AddressSpace, IPv4AddressType, IPv4NetworkType


class LanSwitchInterconnectInterfaceBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="LanSwitchInterconnectInterfaceBlock",
):
    """An inactive :term:`LAN` Switch Interconnect interface."""

    interface_name: str | None = None
    interface_description: str | None = None


class LanSwitchInterconnectInterfaceBlockProvisioning(
    LanSwitchInterconnectInterfaceBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    """A :term:`LAN` Switch Interconnect interface that is being provisioned."""

    interface_name: str
    interface_description: str


class LanSwitchInterconnectInterfaceBlock(
    LanSwitchInterconnectInterfaceBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
):
    """An active Switch Interconnect interface."""

    interface_name: str
    interface_description: str


class LanSwitchInterconnectRouterSideBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="LanSwitchInterconnectRouterSideBlock",
):
    """An inactive :term:`LAN` Switch Interconnect router side."""

    node: RouterBlockInactive
    ae_iface: str | None = None
    ae_members: LAGMemberList[LanSwitchInterconnectInterfaceBlockInactive]
    ipv4_address: IPv4AddressType | None = None


class LanSwitchInterconnectRouterSideBlockProvisioning(
    LanSwitchInterconnectRouterSideBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    """A :term:`LAN` Switch Interconnect router side that is being provisioned."""

    node: RouterBlockProvisioning
    ae_iface: str | None = None
    ae_members: LAGMemberList[LanSwitchInterconnectInterfaceBlockProvisioning]  # type: ignore[assignment]
    ipv4_address: IPv4AddressType | None


class LanSwitchInterconnectRouterSideBlock(
    LanSwitchInterconnectRouterSideBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
):
    """An active :term:`LAN` Switch Interconnect router side."""

    node: RouterBlock
    ae_iface: str
    ae_members: LAGMemberList[LanSwitchInterconnectInterfaceBlock]  # type: ignore[assignment]
    ipv4_address: IPv4AddressType | None


class LanSwitchInterconnectSwitchSideBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="LanSwitchInterconnectSwitchSideBlock",
):
    """An inactive :term:`LAN` Switch Interconnect switch side."""

    switch: SwitchBlockInactive
    ae_iface: str | None = None
    ae_members: LAGMemberList[LanSwitchInterconnectInterfaceBlockInactive]
    ipv4_address: IPv4AddressType | None = None


class LanSwitchInterconnectSwitchSideBlockProvisioning(
    LanSwitchInterconnectSwitchSideBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    """A :term:`LAN` Switch Interconnect switch side that is being provisioned."""

    switch: SwitchBlockProvisioning
    ae_iface: str | None = None
    ae_members: LAGMemberList[LanSwitchInterconnectInterfaceBlockProvisioning]  # type: ignore[assignment]
    ipv4_address: IPv4AddressType | None


class LanSwitchInterconnectSwitchSideBlock(
    LanSwitchInterconnectSwitchSideBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
):
    """An active :term:`LAN` Switch Interconnect switch side."""

    switch: SwitchBlock
    ae_iface: str
    ae_members: LAGMemberList[LanSwitchInterconnectInterfaceBlock]  # type: ignore[assignment]
    ipv4_address: IPv4AddressType | None


class LanSwitchInterconnectBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="LanSwitchInterconnectBlock",
):
    """A :term:`LAN` Switch Interconnect that's currently inactive, see :class:`LanSwitchInterconnectBlock`."""

    lan_switch_interconnect_description: str | None = None
    lan_switch_interconnect_ip_network: IPv4NetworkType | None = None
    address_space: AddressSpace | None = None
    minimum_links: int | None = None
    router_side: LanSwitchInterconnectRouterSideBlockInactive
    switch_side: LanSwitchInterconnectSwitchSideBlockInactive


class LanSwitchInterconnectBlockProvisioning(
    LanSwitchInterconnectBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    """A :term:`LAN` Switch Interconnect that's currently being provisioned, see :class:`LanSwitchInterconnectBlock`."""

    lan_switch_interconnect_description: str | None = None
    lan_switch_interconnect_ip_network: IPv4NetworkType | None
    address_space: AddressSpace | None = None
    minimum_links: int | None = None
    router_side: LanSwitchInterconnectRouterSideBlockProvisioning
    switch_side: LanSwitchInterconnectSwitchSideBlockProvisioning


class LanSwitchInterconnectBlock(LanSwitchInterconnectBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A :term:`LAN` Switch Interconnect that's currently deployed in the network."""

    #: A human-readable description of this :term:`LAN` Switch Interconnect.
    lan_switch_interconnect_description: str
    #: The :term:`IP` resources for this :term:`VLAN` Switch Interconnect.
    lan_switch_interconnect_ip_network: IPv4NetworkType | None
    #: The address space of the :term:`VLAN` Switch Interconnect. It can be private or public.
    address_space: AddressSpace
    #: The minimum amount of links the :term:`LAN` Switch Interconnect should consist of.
    minimum_links: int
    #: The router side of the :term:`LAN` Switch Interconnect.
    router_side: LanSwitchInterconnectRouterSideBlock
    #: The switch side of the :term:`LAN` Switch Interconnect.
    switch_side: LanSwitchInterconnectSwitchSideBlock
